/*
** Module   : SpawnerImpl.java
** Abstract : spawn the web client.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------Description--------------------------
** 001 MAG 20131118 First version.
*/

package com.goldencode.p2j.main;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.swing.*;
import com.goldencode.p2j.util.*;

/**
 * The spawner implementation.
 * 
 */
public class SpawnerImpl
implements Spawner
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(SpawnerImpl.class.getName());

   /** Remote Object Interface and Synchronization */
   private static final Class<Spawner> SPAWNER_CLASS = Spawner.class;
   
   /** Remote Object notification timeout */
   private static final int TIME_OUT = 5000;
   
   /** Instance of Spawner */
   private static Spawner spawner;
   
   /** Main class */
   private Class<ClientDriver> mainClass = ClientDriver.class;

   /** Embedded server URI string */
   private volatile String uri;
   
   /** Create the singleton instance */
   static 
   {
      spawner = new SpawnerImpl();
   }
   
   /**
    * Get the Spawner instance.
    * 
    * @return  The instance of spawner.
    */
   public static Spawner getInstance()
   {
      return spawner;
   }
   
   /**
    * Register a Remote Object Server.
    */
   private SpawnerImpl()
   {      
      RemoteObject.registerNetworkServer(SPAWNER_CLASS, this);
   }
      
   /**
    * Notify when the remote embedded server is ready and listen.
    * 
    * @param   uri
    *          The server URI on which the remote server listen.
    */   
   @Override
   public void OnRemoteServerListen(String uri)
   {
      synchronized(SPAWNER_CLASS)
      {
         this.uri = uri;
         SPAWNER_CLASS.notifyAll();
      }
   }

   /**
    * Get remote embedded server URI.
    * If no notification is coming and a process already exists kill the process.
    * 
    * @param   shell    
    *          The process instance.
    *
    * @return  The server URI on which the remote server listen
    *          or <code>null</code> if no notification. 
    */
   private String getRemoteServerURI(Process shell)
   {
      String old = uri;
      
      if (uri == null && shell != null)
      {
         shell.destroy();
      }
      
      uri = null;
      return old;
   }
   
   /**
    * Add the terminal options registered into directory. 
    * Default values are defined inside the TerminalOptions class.
    * 
    * @param   command    
    *          Command line tokens list.
    */
   private void addTerminalOptions(List<String> command)
   {
      // default terminal options
      int rows = TerminalOptions.DEF_ROWS;
      int columns = TerminalOptions.DEF_COLUMNS;
      String background = Integer.toHexString(TerminalOptions.DEF_BACKGROUND.getRGB());
      String foreground = Integer.toHexString(TerminalOptions.DEF_FOREGROUND.getRGB());
      String selection = Integer.toHexString(TerminalOptions.DEF_SELECTION.getRGB());
      String fontname = TerminalOptions.DEF_FONTNAME;
      int fontsize = TerminalOptions.DEF_FONTSIZE;

      // get defined terminal options
      rows = Utils.getDirectoryNodeInt(null, "chuiWeb/rows", rows, false);
      columns = Utils.getDirectoryNodeInt(null, "chuiWeb/columns", columns, false);
      background = Utils.getDirectoryNodeString(null, "chuiWeb/background", background, false);
      foreground = Utils.getDirectoryNodeString(null, "chuiWeb/foreground", foreground, false);
      selection = Utils.getDirectoryNodeString(null, "chuiWeb/selection", selection, false);
      fontname = Utils.getDirectoryNodeString(null, "chuiWeb/fontname", fontname, false);
      fontsize = Utils.getDirectoryNodeInt(null, "chuiWeb/fontsize", fontsize, false);

      // add to command
      command.add("client:driver:type=" + OutputManager.CHUI_DRIVER_WEB);
      command.add("client:chui:rows=" + columns);
      command.add("client:chui:columns=" + rows);
      command.add("client:chui:background=" + background);
      command.add("client:chui:foreground=" + foreground);
      command.add("client:chui:fontname=" + fontname);
      command.add("client:chui:fontsize=" + fontsize);
   }

   /**
    * Add server properties and the server SSL context.
    * 
    * @param   userDir
    *          Server working directory.
    * @param   command    
    *          Command line tokens list.
    */
   private void addServerOptions(String userDir, List<String> command)
   {
      // add target server ports
      BootstrapConfig serverConfig = SessionManager.get().config();

      String host = serverConfig.getString("net", "server", "host", "localhost");
      command.add("net:server:host=" + host);

      boolean isSecure = serverConfig.getBoolean("net", "connection", "secure", false);
      boolean isInsecure = serverConfig.getBoolean("net", "connection", "insecure", true);

      if (isInsecure)
      {
         if (getServerConfig(serverConfig, command, "net", "server", "port", -1) == -1)
         {
            getServerConfig(serverConfig, command, "net", "server", "insecure_port", -1);
         }
      }

      if (isSecure)
      {
         getServerConfig(serverConfig, command, "net", "server", "secure_port", -1);
      }
      
      // Export server SSL context
      exportSslContext(serverConfig, command, userDir);
   }

   /**
    * Export server SSL context.
    * 
    * @param   config
    *          Server bootstrap configuration.
    * @param   command    
    *          Command line tokens list.
    * @param   userDir    
    *          Server working directory.
    */
   private void exportSslContext(BootstrapConfig config, List<String> command, String userDir)
   {
      // key store file name
      String keyStore = config.getString("security", "keystore", "filename", null);
      if(keyStore != null)
      {
         command.add(String.format("ssl:keystore:filename=%s%s%s", 
                                   userDir, 
                                   File.separator, 
                                   keyStore));         
      }
      // store password
      String keyStorePassword = config.getString("access", "password", "keystore", null);
      if(keyStorePassword != null)
      {
         command.add(String.format("ssl:password:keystore=%s", keyStorePassword));         
      }
      // manager password
      String keyManagerPassword = config.getString("access", "password", "keyentry", null);
      if(keyManagerPassword != null)
      {
         command.add(String.format("ssl:password:keyentry=%s", keyManagerPassword));         
      }
   }
   
   /**
    * Get a key value form configuration file. 
    * If key exists add the value to builder command.
    * 
    * @param   config
    *          Configuration file.
    * @param   command    
    *          Command line tokens list.
    * @param   cname
    *          Category name.
    * @param   gname
    *          Group name.
    * @param   kname
    *          Key name.
    * @param   def
    *          Key value if exists default value otherwise.
    * 
    * @return  The integer value for key.
    */
   private int getServerConfig(BootstrapConfig config,
                               List<String> command,
                               String cname, 
                               String gname, 
                               String kname,
                               int def)
   {
      int value = config.getInt(cname, gname, kname, def);
      if (value != def)
      {
         command.add(String.format("%s:%s:%s=%d", cname, gname, kname, value));
      }
      return value;
   }

   /**
    * Add Java CLASSPATH and the main class.
    * 
    * @param   command    
    *          Command line tokens list.
    */
   private void addClasspath(List<String> command)
   {
      String msg = "Could not determine the location of the p2j.jar file.";
      try
      {
         URL url = mainClass.getProtectionDomain().getCodeSource().getLocation();
         String classPath = url.toURI().getPath();
         File jarFile = new File(classPath);
         // If the source is in a folder and not jar, than this can cause problems
         if (!jarFile.exists())
         {
            throw new IllegalStateException(msg);
         }
         // Native libraries path
         String libPath = jarFile.getParentFile().getPath();
         command.add(String.format("-Djava.library.path=%s", libPath));
         // Add CLASSPATH
         command.add("-classpath");
         command.add(classPath);
         // Add main class
         command.add(mainClass.getName());
      } 
      catch (URISyntaxException e)
      {
         throw new IllegalStateException(msg);
      }
   }

   /**
    * Add client configuration file if exists.
    * 
    * @param   userDir
    *          Client working directory.
    * @param   command    
    *          Command line tokens list.
    */
   private void addConfigFile(String userDir, List<String> command)
   {
      String configFile = "client.xml";
      String clientConfig = String.format("%s%s%s", userDir, File.separator, configFile);
      if (new File(clientConfig).exists())
      {
         command.add(configFile);
      } 
      else
      {
         LOG.warning(String.format("%s does not exist.", clientConfig));
      }
   }

   /**
    * Sets this process builder's working directory.
    * 
    * @param   userDir
    *          Server working directory.
    */
   private File setUserDir(String userDir)
   {
      File fUserDir = new File(userDir);
      if (!fUserDir.exists())
      {
         String msg = "The user dir '%s' specified does not exist.";
         throw new IllegalStateException(String.format(msg, userDir));
      }
      return fUserDir;
   }

   /**
    * Build the command used to lunch the client.
    * 
    * @param   pb    
    *          ProcessBuilder instance.
    * @param   command    
    *          Command line tokens list.
    */
   private void buildCommand(ProcessBuilder pb, List<String> command)
   {
      // Get working directory
      String userDir = System.getProperty("user.dir");
      pb.directory(setUserDir(userDir));
      // Build command
      command.add("java");
      // Add JVM arguments
      command.add("-Xmx1024m");
      command.add("-XX:MaxPermSize=256m");
      // Add CLASSPATH and main class
      addClasspath(command);
      // Add client configuration file
      addConfigFile(userDir, command);
      // Add server options
      addServerOptions(userDir, command);
      // Add terminal options
      addTerminalOptions(command);
   }

   /**
    * Spawn the ClientDriver as a Java process.
    * 
    * @return  The server URI on which the remote server listen or <code>null</code>.
    */
   public String spawn()
   {
      String remoteUri = null;
      
      synchronized (SPAWNER_CLASS) {
         try
         {
            ProcessBuilder pb = new ProcessBuilder();            
            List<String> command = new LinkedList<String>();
            buildCommand(pb, command);
            pb.command(command);
            Process shell = pb.start();
            // wait for notification
            SPAWNER_CLASS.wait(TIME_OUT);
            // get the notification message
            remoteUri = getRemoteServerURI(shell);
         } 
         catch (Exception e)
         {
            LOG.logp(Level.SEVERE,
                     "SpawnerImpl.spawn()",
                     "",
                     LogHelper.generate("Cannot spawn the process " + e.getMessage()));            
         }
      }

      return remoteUri;
   }
}
